from base64 import b64encode
from io import BytesIO
from copy import deepcopy
from typing import (
    List,
    Union,
    Dict,
    Any,
    Optional,
    ClassVar,
    NoReturn,
)

import aiohttp
from PIL import Image

from .request import RequestConnectionBase
from .parser import Soup, ParseIntranet, get_hidden_tags, get_alert_messages


__all__ = (
    'Intranet',
)


Semester = Optional[str]


class Intranet(RequestConnectionBase):
    HOST: ClassVar[str] = 'https://kbuis.bible.ac.kr'
    URIS: ClassVar[Dict[str, str]] = {
        'login_post': HOST + '/ble_login2.aspx',
        'chapel': HOST + '/StudentMng/SM050.aspx',
        'timetable': HOST + '/GradeMng/GD160.aspx',
        'lecture': HOST + '/GradeMng/GD095.aspx',
        'info_edit': HOST + '/SchoolRegMng/SR010.aspx',
        'info_detail': HOST + '/SchoolRegMng/SR020.aspx',
        'image': HOST + '/SchoolRegMng/SR015.aspx?schNo=',
        'evaluation': HOST + '/GradeMng/GD410.aspx'
    }

    async def _post(self, uri: str, *, data: Optional[Dict[str, str]] = None) -> aiohttp.ClientResponse:
        req = await self.request.post(
            uri=uri,
            ssl=False,
            data=data,
        )

        return req

    async def _get(self, uri: str, *, semester: Semester = None) -> Dict[str, str]:
        data = {}
        req = await self._session.get(uri, ssl=False)
        html = await req.text()

        if semester is not None:
            data['ctl00$ContentPlaceHolder1$cbo_YearHg'] = semester

        data.update(get_hidden_tags(html))

        return data

    async def login(self, user_id: str,
                    user_pw: str,
                    user_type: Union[int, str] = 2,
                    *, extra_data: Optional[Dict[str, str]] = None) -> Union[NoReturn, None]:
        """ If the login fails, AlertError is raised. """
        data = {'Txt_1': user_id,
                'Txt_2': user_pw,
                'use_type': user_type}
        if isinstance(extra_data, dict):
            data.update(extra_data)

        req = await self._post(Intranet.URIS['login_post'], data=data)
        html = await req.text()
        ParseIntranet.parse_login(html)

    async def get_chapel(self, *, semester: Semester = None) -> Dict[str, Any]:
        data = {}

        if semester is not None:
            body: Dict[str, str] = await self._get(Intranet.URIS['chapel'], semester=semester)
            data.update(body)

        req = await self._post(Intranet.URIS['chapel'], data=data)
        html = await req.text()
        soup = Soup(markup=html, features='lxml')
        semesters, selected = ParseIntranet.parse_semester(soup)
        summary = ParseIntranet.parse_chapel_summary(soup)
        thead, tbody = ParseIntranet.parse_table(soup)

        result = {
            'selected': selected,  # type: str
            'selectable': semesters,  # type: List[str]
            'summary': summary,  # type: Dict[str, str]
            'thead': thead,  # type: List[str]
            'tbody': tbody,  # type: List[List[str]]
        }

        return result

    async def get_timetable(self, *, semester: Semester = None) -> Dict[str, Any]:
        data = {}

        if semester is not None:
            body = await self._get(Intranet.URIS['timetable'], semester=semester)
            data.update(body)

        req = await self._post(Intranet.URIS['timetable'], data=data)
        html = await req.text()
        soup = Soup(markup=html, features='lxml')

        semesters, selected = ParseIntranet.parse_semester(soup)
        day_of_week, lecture_lines = ParseIntranet.parse_table(soup)
        timetable = ParseIntranet.parse_timetable(soup, lecture_lines)

        result = {
            'selected': selected,  # type: str
            'selectable': semesters,  # type: List[str]
            'thead': day_of_week,  # type: List[str]
            'tbody': timetable,  # type: List[List[List[str]]]  # timetable[day_of_week[lectures]]]
        }

        return result

    async def get_lectures(self, semester: Semester = None) -> Dict[str, Any]:
        data = {}

        if semester is not None:
            body = await self._get(Intranet.URIS['lecture'], semester=semester)
            body['ctl00$ContentPlaceHolder1$hidActionMode'] = 'S'
            data.update(body)

        req = await self._post(Intranet.URIS['lecture'], data=data)
        html = await req.text()
        soup = Soup(markup=html, features='lxml')
        semesters, selected = ParseIntranet.parse_semester(soup)
        head, body = ParseIntranet.parse_table(soup)

        result = {
            'selected': selected,  # type: str
            'selectable': semesters,  # type: List[str]
            'thead': head,  # type: List[str]
            'tbody': body,  # type: List[List[str]]
        }

        return result

    async def get_extension_timetable(self, *, semester: Semester = None):
        timetable = await self.get_timetable(semester=semester)
        if semester is None:
            semester: str = timetable.get('selected')

        lecture_list = await self.get_lectures(semester=semester)
        result = deepcopy(timetable)

        for i, day_of_week_lectures in enumerate(result['tbody']):
            for j, lecture in enumerate(day_of_week_lectures):
                if lecture[0] in lecture_list['tbody']:
                    result['tbody'][i][j].append(lecture_list['table'][lecture[0]])

        return result

    async def get_student_id(self) -> str:
        req = await self._post(Intranet.URIS['info_edit'])
        html = await req.text()

        return ParseIntranet.parse_student_id(html)

    async def get_student_info_from_edit(self) -> Dict[str, str]:
        req = await self._post(Intranet.URIS['info_edit'])
        html = await req.text()

        return ParseIntranet.parse_info_edit(html)

    async def get_student_info_from_detail(self) -> Dict[str, str]:
        req = await self._post(Intranet.URIS['info_detail'])
        html = await req.text()

        return ParseIntranet.parse_info_detail(html)

    async def get_student_image(self, sid: str) -> Dict[str, Union[str, int]]:
        uri = f'{Intranet.URIS["image"]}{sid}'
        req = await self._post(uri)
        try:
            html = await req.text()
            ParseIntranet.parse_student_image(html)
            return {
                'img': '',  # type: str
                'width': 0,  # type: int
                'height': 0  # type: int
            }
        except UnicodeDecodeError:
            pass

        # image load
        data = await req.read()
        # Image
        img = Image.open(BytesIO(data))
        width, height = img.size
        # convert PNG format
        buffered = BytesIO()
        img.save(buffered, format="PNG")
        # bytes -> base64 string
        base64_img = b64encode(buffered.getvalue()).decode('utf-8')

        result = {
            'img': base64_img,  # type: str
            'width': width,  # type: int
            'height': height  # type: int
        }
        return result

    async def get_lms_master_uri(self) -> str:
        req = await self.request.get(Intranet.URIS['info_edit'], ssl=False)
        html = await req.text()

        alerts = get_alert_messages(html)
        if alerts:
            for alert in alerts:
                if '수업평가를 하셔야' in alert:
                    req = await self.request.get(Intranet.URIS['evaluation'], ssl=False)
                    html = await req.text()
                    break

        return ParseIntranet.parse_lms_master_uri(html)
