from typing import (
    List,
    Dict,
    Any,
    Tuple,
    Union,
    Optional,
    ClassVar,
)

import aiohttp

from .request import RequestConnectionBase
from .exceptions import AlertError
from .parser import Soup, ParseLms


__all__ = (
    'Lms',
)


def convert_semester_intranet_to_lms(intranet_semester: str) -> Union[None, Tuple[str, str]]:
    if not intranet_semester.isdigit():
        raise ValueError('Not a number')

    elif len(intranet_semester) != 5:
        raise ValueError('semester format: YYYYS, (e.g. 20191)')

    year = intranet_semester[:4]
    semester = intranet_semester[4]
    if semester == '1':
        converted = '10'
    elif semester == '2':
        converted = '20'
    elif semester == '3':
        converted = '11'
    elif semester == '4':
        converted = '21'
    else:
        raise ValueError('Semester range is between 1 and 4.')

    return year, converted


def convert_semester_lms_to_intranet(year: str, semester: str) -> str:
    if semester == '10':
        converted = '1'
    elif semester == '20':
        converted = '2'
    elif semester == '11':
        converted = '3'
    elif semester == '21':
        converted = '4'
    else:
        raise ValueError('Semester range is [10, 11, 20, 21]')

    return year + converted


class Lms(RequestConnectionBase):
    HOST: ClassVar[str] = 'https://lms.bible.ac.kr'
    URIS: ClassVar[Dict[str, str]] = {
        'login_post': HOST + '/login/index.php',
        'lecture': HOST + '/local/ubion/user/index.php?lang=ko',
        'attendance': HOST + '/local/ubattendance/my_status.php?lang=ko&id=',
        'info': HOST + '/user/user_edit.php?lang=ko'
    }

    async def login_by_uri(self, uri: str):
        req = await self.request.post(
            uri=uri,
            ssl=False,
        )

        html = await req.text()
        ParseLms.parse_login(html)

    async def login(self, user_id: str
                    , user_pw: str,
                    *, data: Optional[Dict[str, str]] = None) -> None:
        """ If the login fails, AlertError is raised. """
        param = {'username': user_id, 'password': user_pw}
        if data and isinstance(data, dict):
            param.update(data)

        req = await self.request.post(
            uri=Lms.URIS['login_post'],
            ssl=False,
            data=param,
        )

        html = await req.text()
        ParseLms.parse_login(html)

    async def get_lectures(self, *, intranet_semester: Optional[str] = None) -> Dict[str, Any]:
        """ If there are no results, an empty dictionary is returned. """
        query = ''
        if intranet_semester is not None:
            year, semester = convert_semester_intranet_to_lms(intranet_semester)
            query = f'?year={year}&semester={semester}'

        uri = Lms.URIS['lecture'] + query
        req = await self.request.get(
            uri=uri,
            ssl=False,
        )

        html = await req.text()
        years, semesters, selected_year, selected_semester = ParseLms.parse_semester(html)
        selected = convert_semester_lms_to_intranet(selected_year, selected_semester)
        selectable = [convert_semester_lms_to_intranet(year, semester)
                      for year in years
                      for semester in semesters]

        lecture_codes = ParseLms.parse_lectures(html)
        return {
            'selected': selected,
            'selectable': selectable,
            'data': lecture_codes
        }

    async def get_attendance(self, lms_lecture_code: str) -> Dict[str, Any]:
        uri = Lms.URIS['attendance'] + lms_lecture_code

        req = await self.request.get(
            uri=uri,
            ssl=False,
        )
        html = await req.text()
        soup = Soup(markup=html, features='lxml')
        extra = ParseLms.parse_header(soup)
        summary = ParseLms.parse_attendance_summary(soup)
        thead, tbody = ParseLms.parse_attendance(soup)

        result = {
            'summary': summary,  # type: Dict[str, str]
            'thead': thead,  # type: List[str]
            'tbody': tbody,  # type: List[List[str]]
            'extra': extra,  # type: Dict[str, str]
        }

        return result

    async def get_student_id(self) -> str:
        uri = Lms.URIS['info']
        req = await self.request.get(
            uri=uri,
            ssl=False,
        )
        html = await req.text()
        result = ParseLms.parse_student_id(html)
        return result

    async def get_info(self) -> Dict[str, str]:
        uri = Lms.URIS['info']
        req = await self.request.get(
            uri=uri,
            ssl=False,
        )
        html = await req.text()
        soup = Soup(markup=html, features='lxml')

        return {
            'sid': ParseLms.parse_student_id(soup),
            'name': ParseLms.parse_name(soup),
            'major': ParseLms.parse_major(soup),
        }
