from typing import (
    List,
    Dict,
    Sequence,
    Optional,
    ClassVar,
)

import aiohttp

from .request import RequestConnectionBase
from .mileage_param import PAIRS, CHANGE_HISTORY_PARAM, SEARCH_PARAM
from .parser import ParseMileage, get_hidden_tags


__all__ = (
    'Mileage',
    # 'OldMileage',
)


def _key_to_kor(s_savename: str) -> List[str]:
    result = []
    for key in s_savename.split('|'):
        if key:
            if key in PAIRS and PAIRS[key]:
                result.append(PAIRS[key])
            else:
                result.append(key)

    return result


def _set_change_history_param(customer_id: str):
    param = CHANGE_HISTORY_PARAM.copy()
    param['CST_NO'] = str(customer_id)

    return param


def _set_search_param(*, student_id: str = None,
                      customer_id: str = None,
                      phone_number: Sequence = None):
    param = SEARCH_PARAM.copy()

    if student_id is not None:
        param['CST_CARD_NO'] = student_id

    if customer_id is not None:
        param['CST_NO'] = customer_id

    if isinstance(phone_number, Sequence) and len(phone_number) == 3:
        param['HP_NO1'], param['HP_NO2'], param['HP_NO3'] = phone_number

    return param


class Mileage(RequestConnectionBase):
    HOST: ClassVar[str] = 'https://asp.netusys.com'
    URIS: ClassVar[Dict[str, str]] = {
        'login_form': HOST + '/login/login_form.jsp',
        'login_post': HOST + '/login/login_check.jsp',
        'data': HOST + '/ddd.sheetAction',
        'customer_page': HOST + '/master/cust/cust020_cst_info.jsp?CST_NO=',
        'customer_info_search': HOST + '/master/cust/cust020.jsp',
    }
    
    async def _get_login_page(self) -> str:
        req = await self.request.get(uri=Mileage.URIS['login_form'], ssl=False)
        html = await req.text()

        return html

    async def _request_login(self, user_id: str,
                             user_pw: str,
                             *, data: Optional[Dict[str, str]] = None) -> aiohttp.ClientResponse:
        data = data.copy() if data is not None else {}
        data.update(user_id=user_id, user_pwd=user_pw)

        req = await self.request.post(
            uri=Mileage.URIS['login_post'],
            ssl=False,
            data=data,
            headers={'referer': Mileage.URIS['login_form']},
        )

        return req

    async def _get_data(self, data: Optional[Dict[str, str]] = None) -> aiohttp.ClientResponse:
        req = await self.request.post(
            uri=Mileage.URIS['data'],
            ssl=False,
            data=data,
        )

        return req

    async def login(self, admin_id: str, admin_pw: str) -> None:
        """ If the login fails, AlertError is raised. """
        login_page_html = await self._get_login_page()
        form_data = get_hidden_tags(login_page_html)
        login_req = await self._request_login(
            admin_id,
            admin_pw,
            data=form_data
        )
        login_html = await login_req.text()

        ParseMileage.parse_login(login_html)

    async def search(self, *, student_id: Optional[str] = None,
                     customer_id: Optional[str] = None) -> Dict[str, List[str]]:
        param = _set_search_param(student_id=student_id, customer_id=customer_id)

        xml_req = await self._get_data(param)
        xml_data = await xml_req.text()

        head = _key_to_kor(param['S_SAVENAME'])
        body = ParseMileage.parse_data(xml_data)
        return {'thead': head, 'tbody': body}

    async def get_customer_id(self, student_id: str) -> str:
        result = await self.search(student_id=student_id)
        if not result or len(result['tbody']) != 1:
            return ''

        for index, element in enumerate(result['thead']):
            if element == '회원번호':
                return result['tbody'][0][index]
        return ''

    async def get_balance(self, *, customer_id: Optional[str] = None, student_id: Optional[str] = None) -> str:
        result = await self.search(customer_id=customer_id, student_id=student_id)
        if not result or len(result['tbody']) != 1:
            return ''

        for index, element in enumerate(result['thead']):
            if element == '포인트가용':
                return result['tbody'][0][index]
        return ''

    async def get_change_history(self, customer_id: str) -> Dict[str, List[str]]:
        param = _set_change_history_param(customer_id)
        xml_req = await self._get_data(param)
        xml_data = await xml_req.text()

        head = _key_to_kor(param['S_SAVENAME'])
        body = ParseMileage.parse_data(xml_data)

        return {'thead': head, 'tbody': body}


# class OldMileage:
#     with open('settings/old_mileage_info.txt', 'r', encoding='utf-8') as f:
#         _old_mileage_info = json.loads(f.read())
#
#     @classmethod
#     def get_old_purchase_history(cls, student_id: str) -> list:
#         if student_id not in cls._old_mileage_info:
#             return []
#
#         return cls._old_mileage_info[student_id]['purchase_history']
#
#     @classmethod
#     def get_old_change_history(cls, student_id: str) -> list:
#         if student_id not in cls._old_mileage_info:
#             return []
#
#         return cls._old_mileage_info[student_id]['change_history']
#
#     @classmethod
#     def get_balance(cls, student_id: str) -> int:
#         if student_id not in cls._old_mileage_info:
#             return 0
#
#         return int(cls._old_mileage_info[student_id]['AVL_POINT'])
