import re
from functools import wraps
from typing import (
    Dict,
    Tuple,
    Union,
    Sequence,
    List,
    Optional,
    Pattern,
    Callable,
    Any,
    ClassVar,
)

from bs4 import BeautifulSoup
from bs4.element import Tag as bs4_tag

from .exceptions import ParseError, AlertError

__all__ = (
    'Soup',
    'get_soup',
    'get_hidden_tags',
    'get_alert_messages',
    'ParseIntranet',
    'ParseKbu',
    'ParseMileage',
    'ParseLms',
)


class Soup(BeautifulSoup):
    def __init__(self, markup: str = '', features: Optional[str] = None, *args, **kwargs):
        super().__init__(markup, features, *args, **kwargs)
        self.original_markup = markup

    def insert_before(self, successor):
        super().insert_before(successor)

    def insert_after(self, successor):
        super().insert_after(successor)


HTML_OR_SOUP = Union[str, Soup]


def get_soup(html_or_soup: HTML_OR_SOUP, features: str = 'lxml') -> Soup:
    return html_or_soup if isinstance(html_or_soup, Soup) else Soup(markup=html_or_soup, features=features)


def get_hidden_tags(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
    soup = get_soup(html_or_soup)
    hidden_tags = soup.find_all(name='input', type='hidden')
    # If the value-tag of hidden_tags, return empty string.
    return {tag.get('name'): tag.get('value') or '' for tag in hidden_tags}


def get_alert_messages(html_or_soup: HTML_OR_SOUP) -> Optional[List[str]]:
    soup: bs4_tag = get_soup(html_or_soup)

    script_elements = soup.find_all('script')
    if not script_elements:  # can not find a script element
        return None

    pattern = re.compile(r'[^a-zA-Z_]*?alert\s*\((.+?)\)')

    return [message.strip()
            for each in script_elements
            for message in pattern.findall(each.text)]


def _check_alert(alerts: Sequence[str], messages: Sequence[Union[str, Pattern]]) -> str:
    for alert in alerts:
        for message in messages:
            if isinstance(message, Pattern):
                if message.search(alert) is not None:
                    return alert

            elif message in alert:
                return alert

    return ''


def _remove_unexpected_char(text: str):
    return (text.replace('\t', '')
            .replace('\n', '')
            .replace('\r', '')
            .replace(u'\xa0', u''))


def check_blocking_message(*blocking_messages: Union[str, Pattern]) -> Callable:
    def wrapper(func: Callable) -> Callable:
        @wraps(func)
        def check(html_or_soup: HTML_OR_SOUP, *args, is_pass: bool = False, **kwargs) -> Any:
            soup = get_soup(html_or_soup)
            if not is_pass and blocking_messages:
                alerts = get_alert_messages(soup)
                if alerts is not None:
                    alert_message = _check_alert(alerts, blocking_messages)
                    if alert_message:
                        raise AlertError(alert_message)
            return func(soup, *args, **kwargs)
        return check
    return wrapper


class ParseIntranet:
    DEFAULT_BLOCKING_MESSAGES: ClassVar[List[str]] = [
        '세션이 종료되었습니다.',
    ]

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        '존재하지 않는 회원ID 입니다',
        '비밀번호가 일치하지 않습니다',
    )
    def parse_login(html_or_soup: HTML_OR_SOUP) -> None:
        pass

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        '수업평가를 하셔야',
    )
    def parse_semester(html_or_soup: HTML_OR_SOUP) -> Tuple[List[str], str]:
        soup = get_soup(html_or_soup)
        semesters = []
        selected = ''

        select_tag = soup.find('select', attrs={'id': 'ctl00_ContentPlaceHolder1_cbo_YearHg'})
        if not select_tag:
            raise ParseError('Could not find a select tag', soup.original_markup)

        options = select_tag.find_all('option')
        if not options:
            raise ParseError('Could not find option tags', soup.original_markup)

        for option in options:
            semesters.append(option.attrs['value'])
            if 'selected' in option.attrs:
                selected = option.attrs['value']

        return semesters, selected

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        '수업평가를 하셔야',
    )
    def parse_chapel_summary(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)

        tbody: bs4_tag = soup.find('tbody', attrs={'class': 'viewbody'})
        if not tbody:
            raise ParseError('Could not find a tbody of viewbody-class.', soup.original_markup)

        result = {}
        for th_tag, td_tag in zip(tbody.find_all('th'), tbody.find_all('td')):
            if th_tag is None or td_tag is None:
                continue

            td_text = td_tag.text.replace(u'\xa0', u' ')
            match = re.search(r'\d+', td_text)

            if not match:  # if not updated yet
                result[th_tag.text] = ''
            else:
                result[th_tag.text] = match.group()

        return result

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        '수업평가를 하셔야',
    )
    def parse_table(html_or_soup: HTML_OR_SOUP) -> Tuple[List[str], List[List[str]]]:
        soup = get_soup(html_or_soup)

        thead: bs4_tag = soup.find('thead', attrs={'class': 'mhead'})
        tbody: bs4_tag = soup.find('tbody', attrs={'class': 'mbody'})

        if not thead:
            raise ParseError('Could not find a thead of mhead-class.', soup.original_markup)
        elif not tbody:
            raise ParseError('Could not find a tbody of mbody-class.', soup.original_markup)

        head = [th.text for th in thead.find_all('th')]
        body = [
            [td.text for td in tr_tags.find_all('td')]
            for tr_tags in tbody.find_all('tr')
        ]
        return head, body

    @staticmethod
    @check_blocking_message()
    def parse_timetable(html_or_soup: HTML_OR_SOUP, scrambled_table: List[List[str]]) -> List[List[List[str]]]:
        timetable = [[] for _ in range(5)]  # Mon ~ Fri

        for lectures in scrambled_table:
            for i, lecture_info in enumerate(lectures):
                lecture_info = lecture_info.strip()
                if not lecture_info:  # empty cell
                    continue

                if re.search(r'\)\d+:', lecture_info):  # if a room is assigned
                    match = re.match(r'(.+)\((.*)\)(\d{2}:\d{2})\s*~\s*(\d{2}:\d{2})', lecture_info)
                else:  # if a room is not assigned, the second capture is an empty string ''.
                    match = re.match(r'(.+)()(\d{2}:\d{2})\s*~\s*(\d{2}:\d{2})', lecture_info)

                if not match:
                    raise ParseError('Could not find timetable information.', get_soup(html_or_soup).original_markup)

                lecture_detail = [
                    match.group(1),  # subject
                    match.group(2),  # room
                    match.group(3),  # start_time
                    match.group(4),  # end_time
                ]
                timetable[i].append(lecture_detail)
        return timetable

    @staticmethod
    @check_blocking_message(
        '수업평가를 하셔야',
    )
    def parse_student_id(html_or_soup: HTML_OR_SOUP) -> str:
        soup = get_soup(html_or_soup)
        student_tag: bs4_tag = soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_1"})

        if not student_tag:
            raise ParseError('Could not find a student-id', soup.original_markup)

        return student_tag.text

    @staticmethod
    @check_blocking_message(
        '수업평가를 하셔야',
    )
    def parse_info_edit(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)

        tags = {
            'sid': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_1"}),
            'name': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_4"}),
            # 'gender': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_3"}),
            # 'phone': soup.find("input", {"id": "ctl00_ContentPlaceHolder1_Txt_4"}),
            # 'email': soup.find("input", {"id": "ctl00_ContentPlaceHolder1_Txt_12"}),
        }

        result_tags = {}
        for key, tag in tags.items():
            if not tag:
                raise ParseError(f'Could not find a {key}', soup.original_markup)
            result_tags[key] = tag.text or tag.get('value')

        return result_tags

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        '수업평가를 하셔야',
    )
    def parse_info_detail(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)

        tags = {
            'major': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_2"}),
            'grade': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_3"}),
            'status': soup.find("span", {"id": "ctl00_ContentPlaceHolder1_Lab_6"}),
        }

        result_tags = {}
        for key, tag in tags.items():
            if not tag:
                raise ParseError(f'Could not find a {key}', soup.original_markup)
            result_tags[key] = tag.text

        return result_tags

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
    )
    def parse_student_image(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        pass

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
    )
    def parse_lms_master_uri(html_or_soup: HTML_OR_SOUP) -> str:
        soup = get_soup(html_or_soup)
        link = soup.find('a', attrs={'title': '학습관리시스템(LMS) 바로가기'}, href=True)
        return link.get('href') or ''


class ParseKbu:
    @staticmethod
    def parse_notice(html_or_soup: HTML_OR_SOUP):
        soup = get_soup(html_or_soup)

        table: bs4_tag = soup.find('table', attrs={'id': 'GridView1'})
        if not table:
            raise ParseError('Could not find a table', soup.original_markup)

        th_tags = table.find_all('th')
        if not th_tags:
            raise ParseError('Could not find th-tags', soup.original_markup)

        keys = [_remove_unexpected_char(th.text) for th in th_tags]
        keys.append('링크')

        tr_tags = table.find_all('tr', attrs={'class': ''})
        if not tr_tags:
            raise ParseError('Could not find tr-tags', soup.original_markup)

        values = []
        for tr in tr_tags[1:-1]:  # th와 페이지번호 제거
            td_tags = tr.find_all('td')
            if td_tags:
                row = [_remove_unexpected_char(td.text) for td in td_tags]
                href = tr.find('a', href=True)
                row.append('http://www.bible.ac.kr/bible_report/' + href['href'] if href else '')
                values.append(row)

        return {
            'thead': keys,
            'tbody': values
        }

    @staticmethod
    def parse_notice_page(html_or_soup: HTML_OR_SOUP):
        soup = get_soup(html_or_soup)

        title = soup.find('div', attrs={'class': 'borad_title'})
        writer = soup.find('span', attrs={'id': 'lblWriter'})
        date = soup.find('span', attrs={'id': 'lblReg_date'})

        if not title:
            raise ParseError('Could not find a title', soup.original_markup)
        elif not writer:
            raise ParseError('Could not find a writer', soup.original_markup)
        elif not date:
            raise ParseError('Could not find a date', soup.original_markup)

        return {
            'title': title.text,
            'writer': writer.text,
            'date': date.text,
        }


class ParseMileage:
    DEFAULT_BLOCKING_MESSAGES: ClassVar[List[str]] = [

    ]

    @staticmethod
    def split_parameter(s_savename: str) -> List[str]:
        """ Convert str to list

        :param s_savename: S_SAVENAME e.g. '|SALE_DATE|SHOP_NM|PROD_NM|SALE_QTY|CST_USE_POINT|'
        :return: list e.g. ['SALE_DATE', 'SHOP_NM', 'PROD_NM', 'SALE_QTY', 'CST_USE_POINT']
        """
        return [param for param in s_savename.split('|') if param]

    @staticmethod
    @check_blocking_message(
        *DEFAULT_BLOCKING_MESSAGES,
        'ID/PW가 일치하지 않습니다.',
        '오류가 발생하였습니다.',
    )
    def parse_login(html: str) -> Tuple[bool, str]:
        pass

    @staticmethod
    def parse_data(xml_data: str) -> List[List[str]]:
        """
        :param xml_data: html responded after get_data()
        :return: [['20191121', '성서대.마트', '광동)제주삼다수 500ml', '1', '0'], ...] or
        [['20191121', '20191121', '5', '-5000', '성서대.마트 : 포인트 사용(승인)'], ... ]
        """
        if '세션정보가 없음' in xml_data:  # 로그인 세션이 만료되었을 때
            raise ParseError('No session information.', xml_data)
        elif '<DATA' not in xml_data:  # 파라미터가 잘못되었거나, 부족할 때
            raise ParseError('The data does not exist.', xml_data)
        else:
            result = []

            tr_scopes = re.findall(r'<TR>(.*?)</TR>', xml_data, flags=re.DOTALL)
            for tr_scope in tr_scopes:
                td_scope = re.findall(r'<TD><!\[CDATA\[(.*?)\]\]></TD>', tr_scope, flags=re.DOTALL)
                if td_scope:
                    result.append([value for value in td_scope[1:]])

            return result


class ParseLms:
    @staticmethod
    def _check_blocking(text: str) -> bool:
        if ('이 강좌에 등록할 수 없습니다.' in text
                or '데이터베이스 테이블 course 에서 데이터 레코드를 찾지 못 함' in text):
            return True
        return False

    @staticmethod
    @check_blocking_message(
        'LMS에 존재하지 않는',
    )
    def parse_login(html_or_soup: HTML_OR_SOUP) -> None:
        soup = get_soup(html_or_soup)
        error_tag: bs4_tag = soup.find('p', attrs={'class': 'error_message text-danger'})
        if error_tag is not None:
            raise AlertError(error_tag.text)

    @staticmethod
    def parse_semester(html_or_soup: HTML_OR_SOUP) -> Tuple[List[str], List[str], str, str]:
        soup = get_soup(html_or_soup)

        year_select: bs4_tag = soup.find('select', attrs={'id': 'year'})
        semester_select: bs4_tag = soup.find('select', attrs={'id': 'semester'})

        if not year_select or not semester_select:
            raise ParseError('Could not find a select tag', soup.original_markup)

        years = []
        semesters = []
        selected_year = ''
        selected_semester = ''

        for option in year_select.find_all('option'):
            years.append(option.attrs['value'])
            if 'selected' in option.attrs:
                selected_year = option.attrs['value']

        for option in semester_select.find_all('option'):
            semesters.append(option.attrs['value'])
            if 'selected' in option.attrs:
                selected_semester = option.attrs['value']

        return years, semesters, selected_year, selected_semester

    @staticmethod
    def parse_lectures(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)
        labels = soup.find_all('div', attrs={'class': 'course_label_re_02'})
        if labels is None:
            raise ParseError('Could not find a labels.', soup.original_markup)

        result = {}
        for label in labels:
            row = label.find('a', href=True)
            if not row:
                continue

            matching_code = re.search(r"[?&]id=(\d+)", row['href'])
            if not matching_code:
                continue

            lecture_name = re.sub(r"\[.*?\]", '', row.text)
            lms_lecture_code = matching_code.group(1)
            result[lecture_name] = lms_lecture_code

        return result

    @staticmethod
    def parse_header(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)

        if ParseLms._check_blocking(soup.text):  # 불러올 수 없는 경우
            return {}

        header_tag: bs4_tag = soup.find('div', attrs={'class': 'attendance_book_list'})
        v: bs4_tag

        result = {}
        for each in header_tag.find_all('li'):
            key, value = _remove_unexpected_char(each.text).split(':', maxsplit=1)
            result[key.strip()] = value.strip()

        return result

    @staticmethod
    def parse_attendance(html_or_soup: HTML_OR_SOUP) -> Tuple[List[str], List[List[str]]]:
        soup = get_soup(html_or_soup)

        if ParseLms._check_blocking(soup.text):  # 불러올 수 없는 경우
            return [], [[]]

        table_tag: bs4_tag = soup.find('table', attrs={'class': 'attendance_my table table-bordered'})
        if not table_tag:
            raise ParseError('Could not find a table.', soup.original_markup)

        thead: bs4_tag = table_tag.find('thead')
        tbody: bs4_tag = table_tag.find('tbody')

        if not thead:
            raise ParseError('Could not find a thead.', soup.original_markup)

        elif not tbody:
            raise ParseError('Could not find a tbody.', soup.original_markup)

        else:
            head = [th.text for th in thead.find_all('th')]
            body = [
                [
                    _remove_unexpected_char(td.text)
                    for td in tr_tags.find_all('td')
                ]
                for tr_tags in tbody.find_all('tr')
            ]

            return head, body

    @staticmethod
    def parse_attendance_summary(html_or_soup: HTML_OR_SOUP) -> Dict[str, str]:
        soup = get_soup(html_or_soup)

        if ParseLms._check_blocking(soup.text):  # 불러올 수 없는 경우
            return {}

        table_tag: bs4_tag = soup.find('table', attrs={'class': 'attendance_my table table-bordered'})

        if not table_tag:
            raise ParseError('Could not find a table.', soup.original_markup)

        tfoot: bs4_tag = table_tag.find('tfoot')

        if not tfoot:
            raise ParseError('Could not find a tfoot.', soup.original_markup)

        else:
            td_tag = tfoot.find('td')
            if not td_tag:
                raise ParseError('Could not find a td tag.', soup.original_markup)

            result = {}
            for each in td_tag.text.split(','):
                key, value = each.split(':')
                key = _remove_unexpected_char(key.strip())
                value = _remove_unexpected_char(value.strip())
                result[key] = value

            return result

    @staticmethod
    def parse_student_id(html_or_soup: HTML_OR_SOUP) -> str:
        soup = get_soup(html_or_soup)

        sid_container = soup.find('div', attrs={'id': 'fitem_id_idnumber'})
        if not sid_container:
            raise ParseError('Could not find a sid_container', soup.original_markup)

        sid_tag: bs4_tag = sid_container.find('div', attrs={'class', 'felement fstatic'})
        if not sid_tag:
            raise ParseError('Could not find a sid', soup.original_markup)

        return sid_tag.get_text().strip()

    @staticmethod
    def parse_name(html_or_soup: HTML_OR_SOUP) -> str:
        soup = get_soup(html_or_soup)

        name_container = soup.find('div', attrs={'id': 'fitem_id_firstname'})
        if not name_container:
            raise ParseError('Could not find a name_container', soup.original_markup)

        name_tag: bs4_tag = name_container.find('input')
        if not name_tag:
            raise ParseError('Could not find a name', soup.original_markup)

        name = name_tag.get('value') or ''
        return name.strip()

    @staticmethod
    def parse_major(html_or_soup: HTML_OR_SOUP) -> str:
        soup = get_soup(html_or_soup)

        major_container = soup.find('div', attrs={'id': 'fitem_id_department'})
        if not major_container:
            raise ParseError('Could not find a name_container', soup.original_markup)

        major_tag: bs4_tag = major_container.find('input')
        if not major_tag:
            raise ParseError('Could not find a name', soup.original_markup)

        major = major_tag.get('value') or ''
        return major.strip()