import asyncio
import pickle
from base64 import b64encode, b64decode
from typing import (
    Dict,
    Optional,
)

import aiohttp

from .exceptions import *

__all__ = (
    'RequestConnectionBase',
    'Request',
)


class RequestConnectionBase:
    def __init__(self, *, headers: Optional[Dict[str, str]] = None,
                 session: Optional[aiohttp.ClientSession] = None,
                 timeout_sec: float = 20.0,
                 cookie_str: Optional[str] = None,
                 **kwargs):
        if not isinstance(session, aiohttp.ClientSession):
            session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(timeout_sec),
                headers=headers,
                **kwargs
            )
        self._session = session
        self.request = Request(self._session)
        if cookie_str is not None:
            self.import_cookie(cookie_str)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        """ Release all acquired resources. """
        await self._session.close()

    def export_cookie(self) -> str:
        try:
            cookie_byte = pickle.dumps(self._session._cookie_jar._cookies, pickle.HIGHEST_PROTOCOL)
            return b64encode(cookie_byte).decode('ascii')
        except TypeError:
            return ''

    def import_cookie(self, cookie_str: str) -> None:
        self._session._cookie_jar._cookies = pickle.loads(b64decode(cookie_str))


class Request:
    def __init__(self, session: aiohttp.ClientSession):
        self._session: aiohttp.ClientSession = session

    async def get(self, uri: str, *, allow_redirects: bool = True, **kwargs) -> aiohttp.ClientResponse:
        return await self._request('get', uri, allow_redirects=allow_redirects, **kwargs)

    async def post(self, uri: str, *, data: Optional[Dict[str, str]] = None, **kwargs) -> aiohttp.ClientResponse:
        return await self._request('post', uri, data=data, **kwargs)

    async def _request(self, method: str, uri: str, *args, **kwargs) -> aiohttp.ClientResponse:
        try:
            request = getattr(self._session, method)
            req: aiohttp.ClientResponse = await request(uri, *args, **kwargs)
        except aiohttp.ClientConnectorError as e:  # 연결 실패
            raise NetworkConnectionError from e
        except asyncio.TimeoutError as e:
            raise NetworkTimeoutError from e
        else:
            http_status = req.status

            if http_status // 100 != 2:  # http status code : not 2xx
                raise BadHTTPStatus(f'http status code is {http_status}')

            return req
